import datetime
import json
import time  # We monkeypatch this.

from django.conf import settings
from django.contrib.auth import logout
from django.contrib.auth.models import User
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured, MiddlewareNotUsed
from django.forms import ValidationError
from django.http import HttpRequest, HttpResponse, HttpResponseForbidden
from django.test import TestCase
from django.test.utils import override_settings
from django.urls import reverse
from django.utils import timezone

from security.auth import min_length
from security.auth_throttling import Middleware as AuthThrottlingMiddleware
from security.auth_throttling import (attempt_count, default_delay_function,
                                      delay_message, increment_counters,
                                      reset_counters, throttling_delay)
from security.middleware import (BaseMiddleware,
                                 ContentSecurityPolicyMiddleware,
                                 DoNotTrackMiddleware,
                                 MandatoryPasswordChangeMiddleware,
                                 ReferrerPolicyMiddleware,
                                 SessionExpiryPolicyMiddleware,
                                 XFrameOptionsMiddleware)
from security.models import PasswordExpiry
from security.password_expiry import never_expire_password
from security.views import csp_report, require_ajax

try:
    # Python 3
    from unittest.mock import MagicMock
except ImportError:
    # Python 2 requires mock library
    from mock import MagicMock

mocked_custom_logout = MagicMock(side_effect=logout)


def login_user(func):
    """
    A decorator that will create a valid user in the database and
    then log that user in. We expect self to be a DjangoTestCase,
    or some object with a similar interface.
    """

    def wrapper(self, *args, **kwargs):
        username_local = "a2fcf54f63993b7"
        password_local = "d8327deb882cf90"
        email_local = "testuser@example.com"
        user = User.objects.create_user(
            username=username_local,
            email=email_local,
            password=password_local,
        )
        user.is_superuser = True
        user.save()
        PasswordExpiry.objects.create(user=user).never_expire()
        self.client.login(username=username_local, password=password_local)
        func(self, *args, **kwargs)
        self.client.logout()
        user.delete()

    return wrapper


class CustomLoginURLMiddleware(BaseMiddleware):
    """Used to test the custom url support in the login required middleware."""

    def process_request(self, request):
        request.login_url = "/custom-login/"


class BaseMiddlewareTestMiddleware(BaseMiddleware):
    REQUIRED_SETTINGS = ("R1", "R2")
    OPTIONAL_SETTINGS = ("O1", "O2")

    def load_setting(self, setting, value):
        if not hasattr(self, "loaded_settings"):
            self.loaded_settings = {}
        self.loaded_settings[setting] = value

    def process_response(self, request, response):
        response.loaded_settings = self.loaded_settings
        return response

    def process_exception(self, request, exception):
        return self.process_response(request, HttpResponse())


class BaseMiddlewareTests(TestCase):
    def __init__(self, *args, **kwargs):
        super(BaseMiddlewareTests, self).__init__(*args, **kwargs)
        module_name = BaseMiddlewareTests.__module__
        self.MIDDLEWARE_NAME = module_name + ".BaseMiddlewareTestMiddleware"

    def test_settings_initially_loaded(self):
        expected_settings = {"R1": 1, "R2": 2, "O1": 3, "O2": 4}
        with self.settings(MIDDLEWARE=(self.MIDDLEWARE_NAME,), **expected_settings):
            response = self.client.get("/home/")
            self.assertEqual(expected_settings, response.loaded_settings)

    def test_required_settings(self):
        with self.settings(MIDDLEWARE=(self.MIDDLEWARE_NAME,)):
            self.assertRaises(ImproperlyConfigured, self.client.get, "/home/")

    def test_optional_settings(self):
        with self.settings(MIDDLEWARE=(self.MIDDLEWARE_NAME,), R1=True, R2=True):
            response = self.client.get("/home/")
            self.assertEqual(None, response.loaded_settings["O1"])
            self.assertEqual(None, response.loaded_settings["O2"])

    def test_setting_change(self):
        with self.settings(MIDDLEWARE=(self.MIDDLEWARE_NAME,), R1=123, R2=True):
            response = self.client.get("/home/")
            self.assertEqual(123, response.loaded_settings["R1"])

            with override_settings(R1=456):
                response = self.client.get("/home/")
                self.assertEqual(456, response.loaded_settings["R1"])

            response = self.client.get("/home/")
            self.assertEqual(123, response.loaded_settings["R1"])

    def test_load_setting_abstract_method(self):
        base = BaseMiddleware()
        self.assertRaises(NotImplementedError, base.load_setting, None, None)


@override_settings(
    MIDDLEWARE=(
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "security.middleware.LoginRequiredMiddleware",
    )
)
class LoginRequiredMiddlewareTests(TestCase):
    def setUp(self):
        self.login_url = reverse("login")

    def test_aborts_if_auth_middleware_missing(self):
        middleware_classes = settings.MIDDLEWARE
        auth_mw = "django.contrib.auth.middleware.AuthenticationMiddleware"
        middleware_classes = [m for m in middleware_classes if m != auth_mw]
        with self.settings(MIDDLEWARE=middleware_classes):
            self.assertRaises(ImproperlyConfigured, self.client.get, "/home/")

    def test_redirects_unauthenticated_request(self):
        response = self.client.get("/home/")
        self.assertRedirects(response, self.login_url + "?next=/home/")

    def test_redirects_unauthenticated_ajax_request(self):
        response = self.client.get(
            "/home/",
            HTTP_X_REQUESTED_WITH="XMLHttpRequest",
        )
        self.assertEqual(response.status_code, 401)
        self.assertEqual(
            json.loads(response.content.decode("utf-8")),
            {"login_url": self.login_url},
        )

    def test_redirects_to_custom_login_url(self):
        middlware_classes = list(settings.MIDDLEWARE)
        custom_login_middleware = "tests.tests.CustomLoginURLMiddleware"
        with self.settings(
            MIDDLEWARE=[custom_login_middleware] + middlware_classes,
        ):
            response = self.client.get("/home/")
            self.assertRedirects(response, "/custom-login/")
            response = self.client.get(
                "/home/",
                HTTP_X_REQUESTED_WITH="XMLHttpRequest",
            )
            self.assertEqual(response.status_code, 401)
            self.assertEqual(
                json.loads(response.content.decode("utf-8")),
                {"login_url": "/custom-login/"},
            )

    def test_logs_out_inactive_users(self):
        user = User.objects.create_user(
            username="foo",
            password="foo",
            email="a@foo.org",
        )
        never_expire_password(user)
        self.client.login(username="foo", password="foo")
        resp = self.client.get("/home/")
        self.assertEqual(resp.status_code, 200)  # check we are logged in
        user.is_active = False
        user.save()
        resp = self.client.get("/home/")
        self.assertRedirects(resp, self.login_url + "?next=/home/")


@override_settings(
    MIDDLEWARE=(
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "security.middleware.MandatoryPasswordChangeMiddleware",
    )
)
class RequirePasswordChangeTests(TestCase):
    def test_require_password_change(self):
        """
        A brand-new user should have an already-expired password, and therefore
        be redirected to the password change form on any request.
        """
        user = User.objects.create_user(
            username="foo", password="foo", email="foo@foo.com"
        )
        self.client.login(username="foo", password="foo")
        try:
            with self.settings(
                MANDATORY_PASSWORD_CHANGE={"URL_NAME": "change_password"}
            ):
                self.assertRedirects(
                    self.client.get("/home/"),
                    reverse("change_password"),
                )
                never_expire_password(user)
                self.assertEqual(self.client.get("/home/").status_code, 200)
        finally:
            self.client.logout()
            user.delete()

    def test_superuser_password_change(self):
        """
        A superuser can be forced to change their password via settings.
        """
        user = User.objects.create_superuser(
            username="foo", password="foo", email="foo@foo.com"
        )
        self.client.login(username="foo", password="foo")
        with self.settings(MANDATORY_PASSWORD_CHANGE={"URL_NAME": "change_password"}):
            self.assertEqual(self.client.get("/home/").status_code, 200)

        try:
            with self.settings(
                MANDATORY_PASSWORD_CHANGE={
                    "URL_NAME": "change_password",
                    "INCLUDE_SUPERUSERS": True,
                }
            ):
                self.assertRedirects(
                    self.client.get("/home/"),
                    reverse("change_password"),
                )
        finally:
            self.client.logout()
            user.delete()

    def test_dont_redirect_exempt_urls(self):
        user = User.objects.create_user(
            username="foo", password="foo", email="foo@foo.com"
        )
        self.client.login(username="foo", password="foo")

        try:
            with self.settings(
                MANDATORY_PASSWORD_CHANGE={
                    "URL_NAME": "change_password",
                    "EXEMPT_URLS": (r"^test1/$", r"^test2/$"),
                    "EXEMPT_URL_NAMES": ("test3", "test4"),
                }
            ):
                # Redirect pages in general
                self.assertRedirects(
                    self.client.get("/home/"),
                    reverse("change_password"),
                )

                # Don't redirect the password change page itself
                self.assertEqual(
                    self.client.get(reverse("change_password")).status_code,
                    200,
                )

                # Don't redirect exempt urls
                self.assertEqual(self.client.get("/test1/").status_code, 200)
                self.assertEqual(self.client.get("/test2/").status_code, 200)
                self.assertEqual(self.client.get("/test3/").status_code, 200)
                self.assertEqual(self.client.get("/test4/").status_code, 200)
        finally:
            self.client.logout()
            user.delete()

    def test_dont_choke_on_exempt_urls_that_dont_resolve(self):
        user = User.objects.create_user(
            username="foo", password="foo", email="foo@foo.com"
        )
        self.client.login(username="foo", password="foo")

        try:
            with self.settings(
                MANDATORY_PASSWORD_CHANGE={
                    "URL_NAME": "change_password",
                    "EXEMPT_URL_NAMES": ("fake1", "fake2"),
                }
            ):
                # Redirect pages in general
                self.assertRedirects(
                    self.client.get("/home/"),
                    reverse("change_password"),
                )
        finally:
            self.client.logout()
            user.delete()

    def test_raises_improperly_configured(self):
        change = MandatoryPasswordChangeMiddleware()
        self.assertRaises(
            ImproperlyConfigured,
            change.load_setting,
            "MANDATORY_PASSWORD_CHANGE",
            {"EXEMPT_URLS": []},
        )


class DecoratorTest(TestCase):
    """
    Testing the AJAXView decorator.
    """

    def require_ajax_test(self):
        @require_ajax
        def ajax_only_view(request):
            self.assertTrue(request.headers.get("x-requested-with") == "XMLHttpRequest")

        request = HttpRequest()
        response = ajax_only_view(request)
        self.assertTrue(isinstance(response, HttpResponseForbidden))
        request.META["HTTP_X_REQUESTED_WITH"] = "XMLHttpRequest"
        response = ajax_only_view(request)
        self.assertFalse(isinstance(response, HttpResponseForbidden))


@override_settings(
    MIDDLEWARE=(
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "security.middleware.SessionExpiryPolicyMiddleware",
        "security.middleware.LoginRequiredMiddleware",
    )
)
class SessionExpiryTests(TestCase):

    def test_session_variables_are_set(self):
        """
        Verify the session cookie stores the start time and last active time.
        """
        self.client.get("/home/")
        now = timezone.now()

        start_time = SessionExpiryPolicyMiddleware._get_datetime_in_session(
            SessionExpiryPolicyMiddleware.START_TIME_KEY, self.client.session
        )
        last_activity = SessionExpiryPolicyMiddleware._get_datetime_in_session(
            SessionExpiryPolicyMiddleware.LAST_ACTIVITY_KEY, self.client.session
        )

        self.assertTrue(now - start_time < datetime.timedelta(seconds=10))
        self.assertTrue(now - last_activity < datetime.timedelta(seconds=10))

    def session_expiry_test(self, key, expired):
        """
        Verify that expired sessions are cleared from the system. (And that we
        redirect to the login page.)
        """
        self.assertTrue(self.client.get("/home/").status_code, 200)
        session = self.client.session
        SessionExpiryPolicyMiddleware._set_datetime_in_session(key, expired, session)
        session.save()
        response = self.client.get("/home/")
        self.assertRedirects(response, reverse("login") + "?next=/home/")

    @login_user
    def test_session_too_old(self):
        """
        Pretend we are 1 second passed the session age time and make sure out
        session is cleared.
        """
        delta = SessionExpiryPolicyMiddleware().SESSION_COOKIE_AGE + 1
        expired = timezone.now() - datetime.timedelta(seconds=delta)
        self.session_expiry_test(SessionExpiryPolicyMiddleware.START_TIME_KEY, expired)

    @login_user
    def test_session_inactive_too_long(self):
        """
        Pretend we are 1 second passed the session inactivity timeout and make
        sure the session is cleared.
        """
        delta = SessionExpiryPolicyMiddleware().SESSION_INACTIVITY_TIMEOUT + 1
        expired = timezone.now() - datetime.timedelta(seconds=delta)
        self.session_expiry_test(
            SessionExpiryPolicyMiddleware().LAST_ACTIVITY_KEY,
            expired,
        )

    @login_user
    def test_exempted_session_expiry_urls(self):
        delta = SessionExpiryPolicyMiddleware().SESSION_INACTIVITY_TIMEOUT + 1
        expired = timezone.now() - datetime.timedelta(seconds=delta)

        self.assertTrue(self.client.get("/home/").status_code, 200)

        session = self.client.session
        SessionExpiryPolicyMiddleware._set_datetime_in_session(
            SessionExpiryPolicyMiddleware.LAST_ACTIVITY_KEY, expired, session
        )
        session.save()

        exempted_response = self.client.get("/accounts/login/")
        not_exempted_response = self.client.get("/home/")

        self.assertTrue(exempted_response.status_code, 200)
        self.assertRedirects(not_exempted_response, reverse("login") + "?next=/home/")

    @login_user
    def test_custom_logout(self):
        delta = SessionExpiryPolicyMiddleware().SESSION_INACTIVITY_TIMEOUT + 1
        expired = timezone.now() - datetime.timedelta(seconds=delta)
        self.session_expiry_test(
            SessionExpiryPolicyMiddleware().LAST_ACTIVITY_KEY,
            expired,
        )
        assert mocked_custom_logout.called


@override_settings(MIDDLEWARE=("security.middleware.NoConfidentialCachingMiddleware",))
class ConfidentialCachingTests(TestCase):
    def setUp(self):
        self.header_values = {
            "Cache-Control": "no-cache, no-store, max-age=0, must-revalidate",
            "Pragma": "no-cache",
            "Expires": "-1",
        }

    @override_settings(
        NO_CONFIDENTIAL_CACHING={
            "WHITELIST_ON": True,
            "BLACKLIST_ON": False,
            "WHITELIST_REGEXES": ["accounts/login/$"],
            "BLACKLIST_REGEXES": ["accounts/logout/$"],
        }
    )
    def test_whitelisting(self):
        # Get Non Confidential Page
        response = self.client.get("/accounts/login/")
        for header, value in self.header_values.items():
            self.assertNotEqual(response.get(header, None), value)
        # Get Confidential Page
        response = self.client.get("/accounts/logout")
        for header, value in self.header_values.items():
            self.assertEqual(response.get(header, None), value)

    @override_settings(
        NO_CONFIDENTIAL_CACHING={
            "WHITELIST_ON": False,
            "BLACKLIST_ON": True,
            "WHITELIST_REGEXES": ["accounts/login/$"],
            "BLACKLIST_REGEXES": ["accounts/logout/$"],
        }
    )
    def test_blacklisting(self):
        # Get Non Confidential Page
        response = self.client.get("/accounts/login/")
        for header, value in self.header_values.items():
            self.assertNotEqual(response.get(header, None), value)
        # Get Confidential Page
        response = self.client.get("/accounts/logout/")
        for header, value in self.header_values.items():
            self.assertEqual(response.get(header, None), value)


@override_settings(MIDDLEWARE=("security.middleware.XFrameOptionsMiddleware",))
class XFrameOptionsDenyTests(TestCase):

    def test_option_set(self):
        """
        Verify the HTTP Response Header is set.
        """
        response = self.client.get("/accounts/login/")
        self.assertEqual(response["X-Frame-Options"], settings.X_FRAME_OPTIONS)

    def test_exclude_urls(self):
        """
        Verify that pages can be excluded from the X-Frame-Options header.
        """
        response = self.client.get("/home/")
        self.assertEqual(response["X-Frame-Options"], settings.X_FRAME_OPTIONS)
        response = self.client.get("/test1/")
        self.assertNotIn("X-Frame-Options", response)

    def test_improperly_configured(self):
        xframe = XFrameOptionsMiddleware()
        self.assertRaises(
            ImproperlyConfigured,
            xframe.load_setting,
            "X_FRAME_OPTIONS",
            "invalid",
        )

        self.assertRaises(
            ImproperlyConfigured,
            xframe.load_setting,
            "X_FRAME_OPTIONS_EXCLUDE_URLS",
            1,
        )

    @override_settings(X_FRAME_OPTIONS_EXCLUDE_URLS=None)
    def test_default_exclude_urls(self):
        # This URL is excluded in other tests, see settings.py
        response = self.client.get("/test1/")
        self.assertEqual(
            response["X-Frame-Options"],
            settings.X_FRAME_OPTIONS,
        )

    @override_settings(X_FRAME_OPTIONS=None)
    def test_default_xframe_option(self):
        response = self.client.get("/home/")
        self.assertEqual(
            response["X-Frame-Options"],
            "deny",
        )


@override_settings(
    AUTHENTICATION_THROTTLING={
        "DELAY_FUNCTION": lambda x, _: (2 ** (x - 1) if x else 0, 0),
        "LOGIN_URLS_WITH_TEMPLATES": [("accounts/login/", "registration/login.html")],
    },
    MIDDLEWARE=(
        "django.contrib.sessions.middleware.SessionMiddleware",
        "django.contrib.auth.middleware.AuthenticationMiddleware",
        "security.auth_throttling.Middleware",
    ),
)
class AuthenticationThrottlingTests(TestCase):
    def setUp(self):
        # monkey patch time
        self.old_time = time.time
        self.time = 0
        time.time = lambda: self.time
        self.user = User.objects.create_user(
            username="foo", password="foo", email="a@foo.org"
        )

    def tearDown(self):
        time.time = self.old_time

    def attempt(self, password):
        return self.client.post(
            "/accounts/login/", {"username": "foo", "password": password}, follow=True
        )

    def reset(self):
        self.client.logout()
        cache.clear()

    def typo(self):
        self.assertTemplateUsed(self.attempt("bar"), "registration/login.html")

    def _succeed(self):
        self.assertTemplateNotUsed(self.attempt("foo"), "registration/login.html")
        self.reset()

    def _fail(self):
        self.assertTemplateUsed(self.attempt("foo"), "registration/login.html")
        self.reset()

    def set_time(self, t):
        self.time = t

    def test_delay_message(self):
        self.assertEqual("0 seconds", delay_message(0))
        self.assertEqual("1 second", delay_message(0.1))
        self.assertEqual("1 second", delay_message(1))
        self.assertEqual("1 minute", delay_message(31))
        self.assertEqual("1 minute", delay_message(60))
        self.assertEqual("1 minute", delay_message(61))
        self.assertEqual("2 minutes", delay_message(90))
        self.assertEqual("2 minutes", delay_message(120))

    def test_counters(self):
        cache.clear()
        increment_counters(username="foo", ip="127.0.0.1")
        increment_counters(username="foo")
        self.assertEqual(attempt_count("username", "foo"), 2)
        self.assertEqual(attempt_count("ip", "127.0.0.1"), 1)
        self.assertEqual(attempt_count("username", "baz"), 0)
        reset_counters(username="foo", ip="127.0.0.1")
        self.assertEqual(attempt_count("username", "foo"), 0)
        self.assertEqual(attempt_count("ip", "127.0.0.1"), 0)
        cache.clear()

    def test_default_delay_function(self):
        """
        The default function will only delay by looking at the username,
        and shouldn't care about ip.
        """
        delay = default_delay_function

        # 100 repeated IPs doesn't result in a delay.
        self.assertEqual(delay(0, 100), (0, 0))

        # first 3 incorrect attempts with a username will not be delayed.
        for i in range(3):
            self.assertEqual(delay(i, 0), (0, 0))

        # forth, fifth, sixth attempts are throttled
        for i in range(4, 7):
            self.assertEqual(delay(i, 0), (5 * 2 ** (i - 3), 0))

        # we max out at 24 hours
        self.assertEqual(delay(100, 0), (24 * 60 * 60, 0))

    def test_per_account_throttling(self):
        """
        Tests that multiple attempts on the same account are throttled
        according to settings.AUTHENTICATION_THROTTLING.
        """
        self.set_time(0)
        self._succeed()

        self.set_time(0)
        self.typo()
        self._fail()

        self.set_time(0)
        self.typo()
        self.set_time(1)
        self._succeed()

        self.set_time(0)
        self.typo()
        self.set_time(1)
        self.typo()
        self.set_time(2)
        self._fail()

        self.set_time(0)
        self.typo()
        self.set_time(1)
        self.typo()
        self.set_time(3)
        self._succeed()

    @override_settings(
        AUTHENTICATION_THROTTLING={
            "DELAY_FUNCTION": lambda x, y: (x, y),
            "LOGIN_URLS_WITH_TEMPLATES": [("accounts/login/", None)],
        }
    )
    def test_too_many_requests_error_when_no_template_provided(self):
        """
        Verify we simply return a 429 error when there is no login template
        provided for us to report an error within.
        """
        cache.clear()

        # first bad attempt
        self.typo()

        # second attempt is throttled as per our delay function
        response = self.attempt("bar")
        self.assertEqual(
            response.status_code,
            429,
            "Expected TooManyRequests Error.",
        )

        cache.clear()

    def test_reset_button(self):
        """
        Tests that the account lockout reset button in the admin interface
        actually works.
        """
        self.set_time(0)
        self.typo()
        admin = User.objects.create_user(
            username="bar", password="bar", email="a@bar.org"
        )
        admin.is_superuser = True
        admin.save()
        self.client.login(username="bar", password="bar")
        self.client.post(
            reverse("reset_username_throttle", args=[self.user.id]),
        )
        self.client.logout()
        self._succeed()

    @override_settings(
        AUTHENTICATION_THROTTLING={
            "DELAY_FUNCTION": lambda x, y: (x, y),
        }
    )
    def test_improperly_configured_middleware(self):
        self.assertRaises(ImproperlyConfigured, AuthThrottlingMiddleware)

    def test_throttle_reset_404_on_unauthorized(self):
        resp = self.client.post(
            reverse("reset_username_throttle", args=[self.user.id]),
        )

        self.assertEqual(resp.status_code, 404)

    def test_throttle_reset_404_on_not_found(self):
        admin = User.objects.create_user(
            username="bar",
            password="bar",
            email="a@bar.org",
        )
        admin.is_superuser = True
        admin.save()
        self.client.login(username="bar", password="bar")
        resp = self.client.post(
            reverse("reset_username_throttle", args=[999]),
        )

        self.assertEqual(resp.status_code, 404)


class AuthTests(TestCase):
    def test_min_length(self):
        self.assertRaises(ValidationError, min_length(6), "abcde")
        min_length(6)("abcdef")


@override_settings(MIDDLEWARE=("security.middleware.ContentSecurityPolicyMiddleware",))
class ContentSecurityPolicyTests(TestCase):
    class FakeHttpRequest(object):
        method = "POST"
        body = (
            """{
          "csp-report": {
            "document-uri": "http://example.org/page.html",
            "referrer": "http://evil.example.com/haxor.html",
            "blocked-uri": "http://evil.example.com/image.png",
            "violated-directive": "default-src 'self'",
            "original-policy": "%s"
          }
        }
        """
            % settings.CSP_STRING
        )
        META = {
            "CONTENT_TYPE": "application/json",
            "REMOTE_ADDR": "127.0.0.1",
            "HTTP_USER_AGENT": "FakeHTTPRequest",
        }

    def test_option_set(self):
        """
        Verify the HTTP Response Header is set.
        """
        response = self.client.get("/accounts/login/")
        self.assertEqual(
            response["Content-Security-Policy"],
            settings.CSP_STRING,
        )

    def test_json(self):

        req = ContentSecurityPolicyTests.FakeHttpRequest()

        parsed = json.loads(req.body)

        self.assertNotEqual(len(parsed), 0)

    # http://www.w3.org/TR/CSP/#sample-violation-report
    def test_csp_view(self):

        req = ContentSecurityPolicyTests.FakeHttpRequest()

        # call the view
        resp = csp_report(req)

        self.assertEqual(resp.status_code, 204)

    def test_csp_gen_1(self):

        csp_dict = {
            "default-src": ["self", "cdn.example.com"],
            "script-src": ["self", "js.example.com"],
            "style-src": ["self", "css.example.com"],
            "img-src": ["self", "img.example.com"],
            "connect-src": [
                "self",
            ],
            "font-src": [
                "fonts.example.com",
            ],
            "object-src": ["self"],
            "media-src": [
                "media.example.com",
            ],
            "frame-src": [
                "*",
            ],
            "sandbox": [
                "",
            ],
            "reflected-xss": "filter",
            "referrer": "origin",
            "report-uri": "http://example.com/csp-report",
        }

        expected = (
            "script-src 'self' js.example.com;"
            "default-src 'self' cdn.example.com;"
            "img-src 'self' img.example.com;"
            "connect-src 'self';"
            "reflected-xss filter;"
            "style-src 'self' css.example.com;"
            "report-uri http://example.com/csp-report;"
            "frame-src *;"
            "sandbox ;"
            "object-src 'self';"
            "media-src media.example.com;"
            "referrer origin;"
            "font-src fonts.example.com"
        )

        csp = ContentSecurityPolicyMiddleware()
        generated = csp._csp_builder(csp_dict)

        # We can't assume the iteration order on the csp_dict, so we split the
        # output, sort, and ensure we got all the results back, regardless of
        # the order.
        expected_list = sorted(x.strip() for x in expected.split(";"))
        generated_list = sorted(x.strip() for x in generated.split(";"))

        self.assertEqual(generated_list, expected_list)

    def test_csp_gen_2(self):
        csp_dict = {"default-src": ("none",), "script-src": ["none"]}
        expected = "default-src 'none'; script-src 'none'"

        csp = ContentSecurityPolicyMiddleware()
        generated = csp._csp_builder(csp_dict)

        expected_list = sorted(x.strip() for x in expected.split(";"))
        generated_list = sorted(x.strip() for x in generated.split(";"))

        self.assertEqual(generated_list, expected_list)

    def test_csp_gen_3(self):
        csp_dict = {
            "script-src": [
                "self",
                "www.google-analytics.com",
                "ajax.googleapis.com",
            ],
        }

        expected = "script-src " "'self' www.google-analytics.com ajax.googleapis.com"

        csp = ContentSecurityPolicyMiddleware()
        generated = csp._csp_builder(csp_dict)

        self.assertEqual(generated, expected)

    def test_csp_gen_err(self):
        # argument not passed as array, expect failure
        csp_dict = {"default-src": "self"}

        csp = ContentSecurityPolicyMiddleware()
        self.assertRaises(MiddlewareNotUsed, csp._csp_builder, csp_dict)

    def test_csp_gen_err2(self):
        csp_dict = {"invalid": "self"}  # invalid directive

        csp = ContentSecurityPolicyMiddleware()
        self.assertRaises(MiddlewareNotUsed, csp._csp_builder, csp_dict)

    def test_csp_gen_err3(self):
        csp_dict = {"sandbox": "none"}  # not a list or tuple, expect failure

        csp = ContentSecurityPolicyMiddleware()
        self.assertRaises(MiddlewareNotUsed, csp._csp_builder, csp_dict)

    def test_csp_gen_err4(self):
        # Not an allowed directive, expect failure
        csp_dict = {"sandbox": ("invalid",)}

        csp = ContentSecurityPolicyMiddleware()
        self.assertRaises(MiddlewareNotUsed, csp._csp_builder, csp_dict)

    def test_csp_gen_err5(self):
        # Not an allowed directive, expect failure
        csp_dict = {"referrer": "invalid"}

        csp = ContentSecurityPolicyMiddleware()
        self.assertRaises(MiddlewareNotUsed, csp._csp_builder, csp_dict)

    def test_csp_gen_err6(self):
        # Not an allowed directive, expect failure
        csp_dict = {"reflected-xss": "invalid"}

        csp = ContentSecurityPolicyMiddleware()
        self.assertRaises(MiddlewareNotUsed, csp._csp_builder, csp_dict)

    def test_enforced_by_default(self):
        with self.settings(CSP_MODE=None):
            response = self.client.get(settings.LOGIN_URL)
            self.assertIn("Content-Security-Policy", response)
            self.assertNotIn("Content-Security-Policy-Report-Only", response)

    def test_enforced_when_on(self):
        with self.settings(CSP_MODE="enforce"):
            response = self.client.get(settings.LOGIN_URL)
            self.assertIn("Content-Security-Policy", response)
            self.assertNotIn("Content-Security-Policy-Report-Only", response)

    def test_report_only_set(self):
        with self.settings(CSP_MODE="report-only"):
            response = self.client.get(settings.LOGIN_URL)
            self.assertNotIn("Content-Security-Policy", response)
            self.assertIn("Content-Security-Policy-Report-Only", response)

    def test_both_enforce_and_report_only(self):
        with self.settings(CSP_MODE="enforce-and-report-only"):
            response = self.client.get(settings.LOGIN_URL)
            self.assertIn("Content-Security-Policy", response)
            self.assertIn("Content-Security-Policy-Report-Only", response)

    def test_invalid_csp_mode(self):
        with self.settings(CSP_MODE="invalid"):
            self.assertRaises(
                MiddlewareNotUsed,
                ContentSecurityPolicyMiddleware,
            )

    def test_no_csp_options_set(self):
        with self.settings(CSP_DICT=None, CSP_STRING=None):
            self.assertRaises(
                MiddlewareNotUsed,
                ContentSecurityPolicyMiddleware,
            )

    def test_both_csp_options_set(self):
        with self.settings(CSP_DICT={"x": "y"}, CSP_STRING="x y;"):
            self.assertRaises(
                MiddlewareNotUsed,
                ContentSecurityPolicyMiddleware,
            )

    def test_sets_from_csp_dict(self):
        with self.settings(
            CSP_DICT={"default-src": ("self",)},
            CSP_STRING=None,
        ):
            response = self.client.get("/accounts/login/")
            self.assertEqual(
                response["Content-Security-Policy"],
                "default-src 'self'",
            )


@override_settings(MIDDLEWARE=("security.middleware.DoNotTrackMiddleware",))
class DoNotTrackTests(TestCase):

    def setUp(self):
        self.request = HttpRequest()
        self.response = HttpResponse()
        self.dnt = DoNotTrackMiddleware(self.response)

    def test_set_DNT_on(self):
        self.request.META["HTTP_DNT"] = "1"
        self.dnt.process_request(self.request)
        self.assertTrue(self.request.dnt)

    def test_set_DNT_off(self):
        self.request.META["HTTP_DNT"] = "off"
        self.dnt.process_request(self.request)
        self.assertFalse(self.request.dnt)

    def test_default_DNT(self):
        self.dnt.process_request(self.request)
        self.assertFalse(self.request.dnt)

    def test_DNT_echo_on(self):
        self.request.META["HTTP_DNT"] = "1"
        self.dnt.process_response(self.request, self.response)
        self.assertIn("DNT", self.response)
        self.assertEqual(self.response["DNT"], "1")

    def test_DNT_echo_off(self):
        self.request.META["HTTP_DNT"] = "off"
        self.dnt.process_response(self.request, self.response)
        self.assertEqual(self.response["DNT"], "off")

    def test_DNT_echo_default(self):
        self.dnt.process_response(self.request, self.response)
        self.assertNotIn("DNT", self.response)


class ReferrerPolicyTests(TestCase):

    def test_option_set(self):
        """
        Verify the HTTP Referrer-Policy Header is set.
        """
        response = self.client.get("/accounts/login/")
        self.assertNotEqual(response["Referrer-Policy"], None)

    def test_default_setting(self):
        with self.settings(REFERRER_POLICY=None):
            response = self.client.get("/accounts/login/")
            self.assertEqual(response["Referrer-Policy"], "same-origin")

    def test_no_referrer_setting(self):
        with self.settings(REFERRER_POLICY="no-referrer"):
            response = self.client.get("/accounts/login/")
            self.assertEqual(response["Referrer-Policy"], "no-referrer")

    def test_no_referrer_when_downgrade_setting(self):
        with self.settings(REFERRER_POLICY="no-referrer-when-downgrade"):
            response = self.client.get("/accounts/login/")
            self.assertEqual(response["Referrer-Policy"], "no-referrer-when-downgrade")

    def test_origin_setting(self):
        with self.settings(REFERRER_POLICY="origin"):
            response = self.client.get("/accounts/login/")
            self.assertEqual(response["Referrer-Policy"], "origin")

    def test_origin_when_cross_origin_setting(self):
        with self.settings(REFERRER_POLICY="origin-when-cross-origin"):
            response = self.client.get("/accounts/login/")
            self.assertEqual(response["Referrer-Policy"], "origin-when-cross-origin")

    def test_same_origin_setting(self):
        with self.settings(REFERRER_POLICY="same-origin"):
            response = self.client.get("/accounts/login/")
            self.assertEqual(response["Referrer-Policy"], "same-origin")

    def test_strict_origin_setting(self):
        with self.settings(REFERRER_POLICY="strict-origin"):
            response = self.client.get("/accounts/login/")
            self.assertEqual(response["Referrer-Policy"], "strict-origin")

    def test_strict_origin_when_cross_origin_setting(self):
        with self.settings(REFERRER_POLICY="strict-origin-when-cross-origin"):
            response = self.client.get("/accounts/login/")
            self.assertEqual(
                response["Referrer-Policy"], "strict-origin-when-cross-origin"
            )

    def test_unsafe_url_setting(self):
        with self.settings(REFERRER_POLICY="unsafe-url"):
            response = self.client.get("/accounts/login/")
            self.assertEqual(response["Referrer-Policy"], "unsafe-url")

    def test_off_setting(self):
        with self.settings(REFERRER_POLICY="off"):
            response = self.client.get("/accounts/login/")
            self.assertEqual("Referrer-Policy" in response, False)

    def test_improper_configuration_raises(self):
        referer_policy_middleware = ReferrerPolicyMiddleware()
        self.assertRaises(
            ImproperlyConfigured,
            referer_policy_middleware.load_setting,
            "REFERRER_POLICY",
            "invalid",
        )


class UnicodeDataTests(TestCase):
    USERNAME = "ñoñó1234"
    IP_ADDRESS = "127.0.0.1"

    def test_unicode_data_in_cache_key(self):
        self._execute_with_data(self.USERNAME, self.IP_ADDRESS)

    def test_types_in_cache_key(self):
        """
        We can send any kind of data for the downstream functions,
        usually strings (maybe the username or email) and ints (maybe the userId)
        """

        self._execute_with_data(1, self.IP_ADDRESS)
        self._execute_with_data(2.67, self.IP_ADDRESS)
        self._execute_with_data(bool, self.IP_ADDRESS)
        self._execute_with_data({"key": "value"}, self.IP_ADDRESS)
        self._execute_with_data([1], self.IP_ADDRESS)
        self._execute_with_data({1, 2}, self.IP_ADDRESS)
        self._execute_with_data((1, 2), self.IP_ADDRESS)
        self._execute_with_data("some_string", self.IP_ADDRESS)

    def _execute_with_data(self, data, ip):
        try:
            increment_counters(key=data, ip=ip)
            reset_counters(key=data, ip=ip)
            throttling_delay(username=data, ip=ip)
            attempt_count(attempt_type="auth", id=data)
        except Exception:
            self.fail("Unicode or incompatible data not allowed")
